import pytest
from datayoga_core import utils
from datayoga_core.blocks.jinja_template.block import Block


@pytest.mark.asyncio
async def test_jinja_template():
    """Test case for applying Jinja template to a field."""
    block = Block({
        "field": "full_name",
        "template": "{{ fname }} {{ lname }}"
    })
    block.init()

    assert await block.run([
        {"fname": "john", "lname": "doe"}
    ]) == utils.all_success([
        {"fname": "john", "lname": "doe", "full_name": "john doe"}
    ])


@pytest.mark.asyncio
async def test_jinja_template_nested_field():
    """Test case for applying Jinja template to a nested field."""
    block = Block({
        "field": "name.full_name",
        "template": "{{ name.fname }} {{ name.lname }}"
    })
    block.init()

    assert await block.run([
        {"name": {"fname": "john", "lname": "doe"}}
    ]) == utils.all_success([
        {"name": {"fname": "john", "lname": "doe", "full_name": "john doe"}}
    ])


@pytest.mark.asyncio
async def test_jinja_template_with_dot():
    """Test case for adding a field with a dot in the name using Jinja template."""
    block = Block({
        "field": "name\.full_name",
        "template": "{{ fname }} {{ lname }}"
    })
    block.init()

    assert await block.run([
        {"fname": "john", "lname": "doe"}
    ]) == utils.all_success([
        {"fname": "john", "lname": "doe", "name.full_name": "john doe"}
    ])


@pytest.mark.asyncio
async def test_jinja_template_missing_field():
    """Test cases for handling missing fields in Jinja template."""
    block = Block({
        "field": "greeting",
        "template": "Hello {{ fname | upper }} {{ lname }}!"
    })
    block.init()

    assert await block.run([
        {"lname": "doe"},
        {"fname": "john"},
        {"fname": "john", "lname": "doe"},
        {}
    ]) == utils.all_success([
        {"lname": "doe", "greeting": "Hello  doe!"},
        {"fname": "john", "greeting": "Hello JOHN !"},
        {"fname": "john", "lname": "doe", "greeting": "Hello JOHN doe!"},
        {"greeting": "Hello  !"}
    ])



import logging
from typing import AsyncGenerator, List, Optional

import sqlalchemy as sa
from datayoga_core import utils
from datayoga_core.blocks.relational import utils as relational_utils
from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer):

    def init(self, context: Optional[Context] = None):
        self.engine, self.db_type = relational_utils.get_engine(
            self.properties["connection"],
            context,
            autocommit=False
        )

        self.schema = self.properties.get("schema")
        self.table = self.properties.get("table")
        self.opcode_field = self.properties.get("opcode_field")
        self.load_strategy = self.properties.get("load_strategy")
        self.keys = self.properties.get("keys")
        self.mapping = self.properties.get("mapping")

        self.tbl = sa.Table(self.table, sa.MetaData(schema=self.schema), autoload_with=self.engine)

        logger.debug(f"Connecting to {self.db_type}")
        self.connection = self.engine.connect()

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        result = self.connection.execution_options(stream_results=True).execute(self.tbl.select())

        while True:
            chunk = result.fetchmany(10000)
            if not chunk:
                break
            for row in chunk:
                yield [utils.add_uid(dict(row._asdict()))]

    def stop(self):
        self.connection.close()
        self.engine.dispose()



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core import utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.properties = utils.format_block_properties(self.properties)

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")
        for row in data:
            for prop in self.properties["fields"]:
                obj = row
                from_field_path = utils.split_field(prop["field"])

                for index, key in enumerate(from_field_path):
                    key = utils.unescape_field(key)
                    if key in obj:
                        if len(from_field_path) == index + 1:
                            del obj[key]
                        else:
                            obj = obj[key]

        return utils.all_success(data)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Any, Dict, List, Optional


@unique
class Status(str, Enum):
    SUCCESS = "SUCCESS"
    REJECTED = "REJECTED"
    FILTERED = "FILTERED"


@dataclass
class Result:
    status: Status
    payload: Optional[Dict[str, Any]] = None
    message: Optional[str] = None


@dataclass
class BlockResult:
    processed: List[Result] = field(default_factory=list)
    filtered: List[Result] = field(default_factory=list)
    rejected: List[Result] = field(default_factory=list)

    def extend(self, other: BlockResult):
        self.processed.extend(other.processed)
        self.filtered.extend(other.filtered)
        self.rejected.extend(other.rejected)

    # allow unpacking
    def __iter__(self):
        return iter((self.processed, self.filtered, self.rejected))


@dataclass
class JobResult(BlockResult):
    # FFU, separate type for JobResult
    pass


SUCCESS = Result(Status.SUCCESS)
FILTERED = Result(Status.FILTERED)



import logging
import sys
from abc import ABCMeta
from typing import Any, Dict, List, Optional

import orjson
from datayoga_core import utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult
from datayoga_core.utils import remove_msg_id

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        for record in data:
            filtered_record = remove_msg_id(record)
            sys.stdout.write(f"{orjson.dumps(filtered_record).decode()}\n")

        # if we made it here, it is a success. return the data and the success result
        return utils.all_success(data)



import logging
from abc import ABCMeta
from contextlib import suppress
from typing import Any, Dict, List, Optional

import sqlalchemy as sa
from datayoga_core import utils, write_utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.blocks.relational import utils as relational_utils
from datayoga_core.context import Context
from datayoga_core.opcode import OpCode
from datayoga_core.result import BlockResult, Result, Status
from sqlalchemy import text
from sqlalchemy.exc import OperationalError
from sqlalchemy.sql.expression import ColumnCollection

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):
    _engine_fields = ("business_key_columns", "mapping_columns", "columns",
                      "delete_stmt", "upsert_stmt", "tbl", "connection", "engine")

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        self.context = context
        self.engine = None
        self.setup_engine()

    def setup_engine(self):
        """Sets up the SQLAlchemy engine and configure it."""
        if self.engine:
            return

        self.engine, self.db_type = relational_utils.get_engine(self.properties["connection"], self.context)
        self.schema = self.properties.get("schema")
        self.table = self.properties.get("table")
        self.opcode_field = self.properties.get("opcode_field")
        self.load_strategy = self.properties.get("load_strategy")
        self.keys = self.properties.get("keys")
        self.mapping = self.properties.get("mapping")
        self.foreach = self.properties.get("foreach")
        self.tbl = sa.Table(self.table, sa.MetaData(schema=self.schema), autoload_with=self.engine)

        if self.opcode_field:
            self.business_key_columns = [column["column"] for column in write_utils.get_column_mapping(self.keys)]
            self.mapping_columns = [column["column"] for column in write_utils.get_column_mapping(self.mapping)]

            self.columns = self.business_key_columns + [x for x in self.mapping_columns
                                                        if x not in self.business_key_columns]

            for column in self.columns:
                if not any(col.name.lower() == column.lower() for col in self.tbl.columns):
                    raise ValueError(f"{column} column does not exist in {self.tbl.fullname} table")

            conditions = []
            for business_key_column in self.business_key_columns:
                for tbl_column in self.tbl.columns:
                    if tbl_column.name.lower() == business_key_column.lower():
                        conditions.append(tbl_column == sa.bindparam(business_key_column))
                        break

            self.delete_stmt = self.tbl.delete().where(sa.and_(*conditions))
            self.upsert_stmt = self.generate_upsert_stmt()

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        """Runs the block with provided data and return the result."""
        logger.debug(f"Running {self.get_block_name()}")
        rejected_records: List[Result] = []

        self.setup_engine()

        if self.opcode_field:
            opcode_groups = write_utils.group_records_by_opcode(data, opcode_field=self.opcode_field)
            # reject any records with unknown or missing Opcode
            for opcode in set(opcode_groups.keys()) - {o.value for o in OpCode}:
                rejected_records.extend([
                    Result(status=Status.REJECTED, payload=record, message=f"unknown opcode '{opcode}'")
                    for record in opcode_groups[opcode]
                ])

            records_to_upsert = opcode_groups[OpCode.CREATE] + opcode_groups[OpCode.UPDATE]
            records_to_delete = opcode_groups[OpCode.DELETE]
            if self.foreach:
                self.execute_upsert(utils.explode_records(records_to_upsert, self.foreach))
                self.execute_delete(utils.explode_records(records_to_delete, self.foreach))
            else:
                self.execute_upsert(records_to_upsert)
                self.execute_delete(records_to_delete)

            return BlockResult(
                processed=[Result(Status.SUCCESS, payload=record)
                           for opcode in OpCode for record in opcode_groups[opcode.value]],
                rejected=rejected_records)
        else:
            logger.debug(f"Inserting {len(data)} record(s) to {self.table} table")
            self.execute(self.tbl.insert(), data)
            return utils.all_success(data)

    def generate_upsert_stmt(self) -> Any:
        """Generates an UPSERT statement based on the DB type"""
        if self.db_type == relational_utils.DbType.PSQL:
            from sqlalchemy.dialects.postgresql import insert

            insert_stmt = insert(self.tbl).values({col: "?" for col in self.columns})
            return insert_stmt.on_conflict_do_update(
                index_elements=self.business_key_columns,
                set_={col: getattr(insert_stmt.excluded, col) for col in self.columns})

        if self.db_type == relational_utils.DbType.MYSQL:
            from sqlalchemy.dialects.mysql import insert

            insert_stmt = insert(self.tbl).values({col: "?" for col in self.columns})
            return insert_stmt.on_duplicate_key_update(ColumnCollection(
                columns=[(x.name, x) for x in [insert_stmt.inserted[column] for column in self.columns]]))

        if self.db_type == relational_utils.DbType.SQLSERVER:
            return sa.sql.text("""
                    MERGE %s AS target
                    USING (VALUES (%s)) AS source (%s) ON (%s)
                    WHEN NOT MATCHED BY target THEN INSERT (%s) VALUES (%s)
                    WHEN MATCHED THEN UPDATE SET %s;
                    """ % (
                relational_utils.construct_table_reference(self.tbl, with_brackets=True),
                ", ".join([f"{sa.bindparam(column)}" for column in self.business_key_columns]),
                ", ".join([f"[{column}]" for column in self.business_key_columns]),
                " AND ".join([f"target.[{column}] = source.[{column}]" for column in self.business_key_columns]),
                ", ".join([f"[{column}]" for column in self.columns]),
                ", ".join([f"{sa.bindparam(column)}" for column in self.columns]),
                ", ".join([f"target.[{column}] = {sa.bindparam(column)}" for column in self.mapping_columns])
            ))

        if self.db_type == relational_utils.DbType.ORACLE:
            return sa.sql.text("""
                    MERGE INTO %s target
                    USING (SELECT 1 FROM DUAL) ON (%s)
                    WHEN NOT MATCHED THEN INSERT (%s) VALUES (%s)
                    WHEN MATCHED THEN UPDATE SET %s
                    """ % (
                relational_utils.construct_table_reference(self.tbl),
                " AND ".join([f"target.{column} = :{column}" for column in self.business_key_columns]),
                ", ".join([f"{column}" for column in self.columns]),
                ", ".join([f"{sa.bindparam(column)}" for column in self.columns]),
                ", ".join([f"target.{column} = {sa.bindparam(column)}" for column in self.mapping_columns])
            ))

        if self.db_type == relational_utils.DbType.DB2:
            return sa.sql.text("""
                MERGE INTO %s AS target
                USING (VALUES (%s)) AS source (%s)
                ON (%s)
                WHEN NOT MATCHED THEN INSERT (%s) VALUES (%s)
                WHEN MATCHED THEN UPDATE SET %s
                """ % (
                relational_utils.construct_table_reference(self.tbl),
                ", ".join([f"{sa.bindparam(column)}" for column in self.business_key_columns]),
                ", ".join([f"{column}" for column in self.business_key_columns]),
                " AND ".join([f"target.{column} = source.{column}" for column in self.business_key_columns]),
                ", ".join([f"{column}" for column in self.columns]),
                ", ".join([f"{sa.bindparam(column)}" for column in self.columns]),
                ", ".join([f"target.{column} = {sa.bindparam(column)}" for column in self.mapping_columns])
            ))

    def execute(self, statement: Any, records: List[Dict[str, Any]]):
        """Executes a SQL statement with given records."""
        if isinstance(statement, str):
            statement = text(statement)

        logger.debug(f"Executing {statement} on {records}")
        connected = False
        try:
            with self.engine.connect() as connection:
                connected = True
                try:
                    connection.execute(statement, records)
                    if not connection._is_autocommit_isolation():
                        connection.commit()
                except OperationalError as e:
                    if self.db_type == relational_utils.DbType.MYSQL:
                        mysql_conn_errors = (
                            2002,  # CR_CONNECTION_ERROR
                            2003,  # CR_CONN_HOST_ERROR
                            2006,  # CR_SERVER_GONE_ERROR
                            2013,  # CR_SERVER_LOST
                            2055,  # CR_SERVER_LOST_EXTENDED
                        )

                        if e.orig.args[0] in mysql_conn_errors:
                            connected = False

                    raise
                except Exception:
                    raise
        except Exception as e:
            if not connected:
                raise ConnectionError(e) from e

            raise

    def execute_upsert(self, records: List[Dict[str, Any]]):
        """Upserts records into the table."""
        if records:
            logger.debug(f"Upserting {len(records)} record(s) to {self.table} table")
            records_to_upsert = []
            for record in records:
                records_to_upsert.append(write_utils.map_record(record, self.keys, self.mapping))

            if records_to_upsert:
                self.execute(self.upsert_stmt, records_to_upsert)

    def execute_delete(self, records: List[Dict[str, Any]]):
        """Deletes records from the table."""
        if records:
            logger.debug(f"Deleting {len(records)} record(s) from {self.table} table")
            records_to_delete = []
            for record in records:
                records_to_delete.append(write_utils.map_record(record, self.keys))

            if records_to_delete:
                self.execute(self.delete_stmt, records_to_delete)

    def stop(self):
        """Disposes of the engine and cleans up resources."""
        with suppress(Exception):
            if self.engine:
                self.engine.dispose()

        for attr in self._engine_fields:
            setattr(self, attr, None)



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core import expression, utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.properties = utils.format_block_properties(self.properties)

        self.fields = {}
        for prop in self.properties["fields"]:
            self.fields[prop["field"]] = expression.compile(prop["language"], prop["expression"])

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")
        for field, expr in self.fields.items():
            expression_results = expr.search_bulk(data)

            for i, row in enumerate(data):
                utils.set_field(row, field, expression_results[i])

        return utils.all_success(data)



import logging
import select
import sys
import uuid
from typing import Any, AsyncGenerator, Dict, List, Optional

import orjson
from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        if select.select([sys.stdin, ], [], [], 0.0)[0]:
            # piped data exists
            for data in sys.stdin:
                for record in self.get_records(data):
                    yield [self.get_message(record)]
        else:
            # interactive mode
            print("Enter data to process:")
            data = input()
            for record in self.get_records(data):
                yield [self.get_message(record)]

    @staticmethod
    def get_records(data: str) -> List[Dict[str, Any]]:
        records = orjson.loads(data)

        if isinstance(records, dict):
            records = [records]

        return records

    def get_message(self, record: Dict[str, Any]) -> Message:
        return {self.MSG_ID_FIELD: str(uuid.uuid4()), **record}



import logging
from typing import Any, Dict, List, Optional

import aiohttp
from datayoga_core import expression, utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.connection import Connection
from datayoga_core.context import Context
from datayoga_core.expression import Expression
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock):
    base_uri: str
    method: str
    request_config: Dict[str, Any]
    response_status_code_field: Optional[str]
    response_headers_field: Optional[str]
    response_content_field: Optional[str]
    timeout: int

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        connection_name = self.properties["connection"]
        connection_details = Connection.get_connection_details(connection_name, context)
        if connection_details["type"] != "http":
            raise ValueError(f"{connection_name} is not an HTTP connection")

        self.base_uri = f"{connection_details['base_uri'].rstrip('/')}"
        self.method = self.properties["method"]
        output = self.properties.get("output", {})
        self.response_status_code_field = output.get("status_code")
        self.response_headers_field = output.get("headers")
        self.response_content_field = output.get("body")
        self.timeout = self.properties.get("timeout", connection_details.get("timeout", 10))

        def process_dict(input_dict, output_dict):
            for key, value in input_dict.items():
                if isinstance(value, dict) and "expression" in value and "language" in value:
                    output_dict[key] = expression.compile(value["language"], value["expression"])
                elif isinstance(value, dict):
                    output_dict[key] = {}
                    process_dict(value, output_dict[key])
                else:
                    output_dict[key] = value

        self.request_config = {}
        process_dict({
            "endpoint": self.properties["endpoint"],
            "headers": {**connection_details.get("headers", {}), **self.properties.get("extra_headers", {})},
            "query_params": {
                **connection_details.get("query_parameters", {}),
                **self.properties.get("extra_query_parameters", {})},
            "payload": self.properties.get("payload", {})
        }, self.request_config)

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")

        block_result = BlockResult()

        def process_dict(input_dict, output_dict):
            for key, value in input_dict.items():
                if isinstance(value, Expression):
                    output_dict[key] = value.search_bulk(data)
                elif isinstance(value, dict):
                    output_dict[key] = {}
                    process_dict(value, output_dict[key])
                else:
                    output_dict[key] = [value] * len(data)

        request_configs = {}
        process_dict(self.request_config, request_configs)

        async with aiohttp.ClientSession() as session:
            for i, row in enumerate(data):
                response = None

                try:
                    url = f"{self.base_uri}/{request_configs['endpoint'][i].lstrip('/')}"
                    headers = {key: value[i] for key, value in request_configs["headers"].items()}
                    query_params = {
                        key: str(value[i]) for key, value in request_configs["query_params"].items()
                        if value[i] is not None}
                    payload = {key: value[i] for key, value in request_configs["payload"].items()}

                    logger.debug(
                        f"Sending HTTP {self.method} request to: {url}\nheaders:{headers}\n\tquery_params: {query_params}\n\tpayload: {payload}")

                    async with session.request(self.method, url, params=query_params, headers=headers, data=payload, timeout=self.timeout) as response:
                        response_status = response.status
                        response_headers = dict(response.headers)
                        response_text = await response.text()

                    logger.debug(f"HTTP response code: {response_status}")
                    logger.debug(f"Response Headers: {response_headers}")
                    logger.debug(f"Response Content: {response_text}")

                    if self.response_status_code_field:
                        utils.set_field(row, self.response_status_code_field, response_status)

                    if self.response_headers_field:
                        utils.set_field(row, self.response_headers_field, response_headers)

                    if self.response_content_field:
                        utils.set_field(row, self.response_content_field, response_text)

                    if response.ok:
                        block_result.processed.append(Result(Status.SUCCESS, payload=row))
                    else:
                        error_message = response_text if response_text else "Unknown error"
                        block_result.rejected.append(
                            Result(
                                status=Status.REJECTED, payload=row,
                                message=f"HTTP request failed with status code {response_status}. Error message: {error_message}"))
                except Exception as e:
                    block_result.rejected.append(
                        Result(status=Status.REJECTED, payload=row, message=f"Error making HTTP request: {f'{e}'}"))

        return block_result



import logging
from typing import Any, Dict, List, Optional

from datayoga_core.context import Context
from datayoga_core.job import Job
from datayoga_core.result import JobResult

logger = logging.getLogger("dy")


def compile(job_settings: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None) -> Job:
    """Compiles a job in YAML.

    Args:
        job_settings (Dict[str, Any]): Job settings.
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Returns:
        Job: Compiled job.
    """
    logger.debug("Compiling job")
    return Job.compile(job_settings, whitelisted_blocks)


def validate(job_settings: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None):
    """Validates a job in YAML.

    Args:
        job_settings (Dict[str, Any]): Job settings.
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Raises:
        ValueError: When the job is invalid.
    """
    logger.debug("Validating job")
    try:
        Job.validate(
            source=job_settings,
            whitelisted_blocks=whitelisted_blocks
        )
    except Exception as e:
        raise ValueError(e)


def transform(
    job_settings: Dict[str, Any],
    data: List[Dict[str, Any]],
    context: Optional[Context] = None,
    whitelisted_blocks: Optional[List[str]] = None
) -> JobResult:
    """Transforms data against a certain job.

    Args:
        job_settings (Dict[str, Any]): Job settings.
        data (List[Dict[str, Any]]): Data to transform.
        context (Optional[Context]): Context. Defaults to None.
        whitelisted_blocks: (Optional[List[str]]): Whitelisted blocks. Defaults to None.

    Returns:
        JobResult: Job result.
    """
    job = compile(job_settings, whitelisted_blocks)
    job.init(context)
    logger.debug("Transforming data")
    return job.transform(data)



import logging

import orjson
import sqlglot

try:
    # older linux doesn't have adequate version
    import pysqlite3 as sqlite3
except ImportError:
    import sqlite3

from abc import abstractmethod
from enum import Enum, unique
from typing import Any, Dict, List, Tuple, Union

import jmespath
from datayoga_core.jmespath_custom_functions import JmespathCustomFunctions

logger = logging.getLogger("dy")


@unique
class Language(str, Enum):
    JMESPATH = "jmespath"
    SQL = "sql"


def get_nested_value(data: Dict[str, Any], key: Tuple[str]) -> Any:
    # get nested key
    for level in key:
        try:
            data = data[level]
        except KeyError:
            # such field does not exist, return None
            return None
    return data


class Expression:
    @abstractmethod
    def compile(self, expression: str):
        """Compiles an expression

        Args:
            expression (str): expression
        """
        raise NotImplementedError

    @abstractmethod
    def search(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Executes the expression on a single data entry

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        raise NotImplementedError

    @abstractmethod
    def search_bulk(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Executes the expression in bulk on a list of data entries

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        raise NotImplementedError

    def filter(self, data: List[Dict[str, Any]], tombstone: bool = False) -> List[Union[Dict[str, Any], None]]:
        """Tests a where clause for an SQL statement

        Args:
            data (List[Dict[str, Any]]): Data
            tombstone: if True, returns None for filtered values

        Returns:
            List[Dict[str, Any]]: Filtered data
        """
        # filter is essentially using a single field returning 0 or 1 for the condition
        if tombstone:
            return [row[0] if row[1] else None for row in zip(data, self.search_bulk(data))]
        else:
            return [row[0] for row in zip(data, self.search_bulk(data)) if row[1]]


class SQLExpression(Expression):
    def compile(self, expression: str):
        # check min sqlite3 version to at least support values clause
        if sqlite3.sqlite_version_info < (3, 8, 8):
            raise ValueError(
                f"must have SQLite v3.8.8 and above to use SQL expressions. Found {sqlite3.sqlite_version_info}")

        # we turn off `check_same_thread` to gain performance benefit by reusing the same connection object
        # safe to use since we are only creating in memory structures
        self.conn = sqlite3.connect(":memory:", check_same_thread=False)
        # we support both single field expressions and multiple fields
        self._is_single_field = True
        try:
            self._fields = orjson.loads(expression)

            # verify this is a dict, and not a list or document ('x' is a valid json)
            if isinstance(self._fields, dict):
                self._is_single_field = False
            else:
                # this is not a dict, treat as a simple expression
                self._fields = {"expr": expression}

        except orjson.JSONDecodeError:
            # this is not a json, treat as a simple expression
            self._fields = {"expr": expression}

        # for each of the expressions, we determine the column names used.
        # they will get parsed and binded for performance instead of traversing the entire payload
        self._column_names = set()
        for _exp in self._fields.values():
            try:
                self._column_names.update(
                    [tuple(column.sql().replace('"', "").split("."))
                     for column in sqlglot.parse_one("SELECT " + _exp.replace('`', '"')).find_all(sqlglot.exp.Column)])
            except Exception:
                # a parse error
                raise ValueError(f"Cannot parse SQL expression: {_exp}")

    def search_bulk(self, data: List[Dict[str, Any]]) -> Any:
        results = self.exec_sql(data, self._fields)
        if self._is_single_field:
            # treat as a simple expression
            return [x.get("expr") for x in results]
        else:
            return results

    def search(self, data: Dict[str, Any]) -> Any:
        return self.search_bulk([data])[0]

    def exec_sql(self, data: List[Dict[str, Any]], expressions: Dict[str, str]) -> List[Dict[str, Any]]:
        """Executes an SQL statement

        Args:
            data (List[Dict[str, Any]]): Data
            expressions: Dict[str, str]: Expressions

        Returns:
            List[Dict[str, Any]]: Query result
        """
        # builds an expression for fetching in memory data
        self.conn.row_factory = sqlite3.Row

        if len(self._column_names) > 0:
            columns_clause = ",".join(f"[column{i+1}] AS `{'.'.join(col)}`" for i, col in enumerate(self._column_names))

            # values in the form of (?,?), (?,?)
            values_clause_row = f"({','.join('?' * len(self._column_names))})"
            values_clause = ",".join([values_clause_row] * len(data))

            subselect = f"SELECT {columns_clause} FROM (VALUES {values_clause})"

            # bind the variables
            data_values = [get_nested_value(row, col) for row in data for col in self._column_names]

            # expressions clause
            expressions_clause = ", ".join(
                [f"{expression} as `{column_name}`" for column_name, expression in expressions.items()])
            # we don't use CTE because of compatibility with older SQLlite versions on Centos7
            statement = f"SELECT {expressions_clause} FROM ({subselect})"

            logger.debug(statement)
            cursor = self.conn.execute(statement, data_values)
            return [dict(x) for x in cursor.fetchall()]

        # a special case where we are only selecting literals. e.g. select current_timstamp or 'x'
        # no need to bind anything. just run it once and copy to all records
        expressions_clause = ", ".join(
            [f"{expression} as `{column_name}`" for column_name, expression in expressions.items()])
        cursor = self.conn.execute(f"select {expressions_clause}")
        return [dict(cursor.fetchone())] * len(data)


class JMESPathExpression(Expression):
    # register custom functions
    options = jmespath.Options(custom_functions=JmespathCustomFunctions())

    def compile(self, expression: str):
        self.expression = jmespath.compile(expression)

    def search(self, data: Dict[str, Any]) -> Any:
        return self.expression.search(data, options=self.options)

    def search_bulk(self, data: List[Dict[str, Any]]) -> Any:
        return [self.search(row) for row in data]


def compile(language: Language, expression: str) -> Expression:
    """Gets a compiled expression class based on the language

    Args:
        language (Language): Language
        expression (str): Expression

    Returns:
        Expression: Expression class
    """
    if language == Language.JMESPATH:
        expression_class = JMESPathExpression()
    elif language == Language.SQL:
        expression_class = SQLExpression()
    else:
        raise ValueError(f"unknown expression language {language}")

    expression_class.compile(expression)
    return expression_class



import pytest
from datayoga_core import utils
from datayoga_core.blocks.rename_field.block import Block


@pytest.mark.asyncio
async def test_rename_existing_field():
    block = Block(
        {
            "from_field": "fname",
            "to_field": "first_name"
        }
    )
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([{"first_name": "john", "lname": "doe"}])


@pytest.mark.asyncio
async def test_rename_missing_field():
    block = Block(
        {
            "from_field": "mname",
            "to_field": "middle_name"
        }
    )
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([
        {"fname": "john", "lname": "doe"}])


@pytest.mark.asyncio
async def test_rename_deep_nested_field():
    block = Block(
        {
            "from_field": "employee.name.fname",
            "to_field": "employee.name.first_name"
        }
    )
    block.init()
    assert await block.run([{"employee": {"name": {"fname": "john", "lname": "doe"}}}]) == utils.all_success([

        {"employee": {"name": {"first_name": "john", "lname": "doe"}}}])


@pytest.mark.asyncio
async def test_rename_nested_field():
    block = Block(
        {
            "from_field": "name.fname",
            "to_field": "name.first_name"
        }
    )
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == utils.all_success([
        {"name": {"first_name": "john", "lname": "doe"}}])


@pytest.mark.asyncio
async def test_rename_nested_field_missing_to_field_parent():
    block = Block(
        {
            "from_field": "name.fname",
            "to_field": "new_name.first_name"
        }
    )
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == utils.all_success([
        {"name": {"lname": "doe"}, "new_name": {"first_name": "john"}}])


@pytest.mark.asyncio
async def test_rename_field_with_dot():
    block = Block(
        {
            "from_field": "name\.fname",
            "to_field": "name\.first_name"
        }
    )
    block.init()
    assert await block.run([{"name.fname": "john", "lname": "doe"}]) == utils.all_success([{"name.first_name": "john", "lname": "doe"}])



import logging
from enum import Enum, unique
from typing import Callable, Tuple

import sqlalchemy as sa
from datayoga_core.connection import Connection
from datayoga_core.context import Context
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import DBAPIConnection
from sqlalchemy.pool import ConnectionPoolEntry

logger = logging.getLogger("dy")


@unique
class DbType(str, Enum):
    DB2 = "db2"
    MYSQL = "mysql"
    ORACLE = "oracle"
    PSQL = "postgresql"
    SQLSERVER = "sqlserver"


DEFAULT_DRIVERS = {
    DbType.DB2: "ibm_db_sa",
    DbType.MYSQL: "mysql+pymysql",
    DbType.ORACLE: "oracle+oracledb",
    DbType.PSQL: "postgresql",
    DbType.SQLSERVER: "mssql+pymssql"
}


def get_engine(connection_name: str, context: Context, autocommit: bool = True) -> Tuple[sa.engine.Engine, DbType]:
    """Creates an SQLAlchemy engine based on the connection details specified in the context.

    Args:
        connection_name (str): The name of the connection to retrieve details for.
        context (Context): The context object providing necessary configurations and settings.
        autocommit (bool, optional): Indicates whether to set the engine's isolation level to autocommit.
            Defaults to True.

    Raises:
        ValueError: If the connection details are invalid or missing required fields.

    Returns:
        Tuple[sa.engine.Engine, DbType]: A tuple containing the created SQLAlchemy engine and the database type.
    """
    connection_details = Connection.get_connection_details(connection_name, context)

    db_type = DbType(connection_details.get("type", "").lower())

    query_args = connection_details.get("query_args", {})
    connect_args = connection_details.get("connect_args", {})
    extra = {}

    if autocommit:
        extra["isolation_level"] = None if db_type == DbType.DB2 else "AUTOCOMMIT"

    if db_type == DbType.ORACLE and connection_details.get("oracle_thick_mode", False):
        lib_dir = connection_details.get("oracle_thick_mode_lib_dir")
        extra["thick_mode"] = {"lib_dir": lib_dir} if lib_dir else {}

    engine = sa.create_engine(
        sa.engine.URL.create(
            drivername=connection_details.get("driver", DEFAULT_DRIVERS.get(db_type)),
            host=connection_details.get("host"),
            port=connection_details.get("port"),
            username=connection_details.get("user"),
            password=connection_details.get("password"),
            database=connection_details.get("database"),
            query=query_args),
        echo=connection_details.get("debug", False),
        connect_args=connect_args,
        pool_pre_ping=True,
        **extra)

    if db_type == DbType.ORACLE:
        # add NLS_DATE_FORMAT for Oracle to explicitly set the date format to ISO
        sa.event.listen(engine, "connect", alter_session_on_oracle_connect)

    if engine.driver == "pymysql":
        # add a listener to configure PyMySQL dialect settings
        sa.event.listen(engine, "connect", create_pymysql_dialect_configuration(engine))
    return engine, db_type


def create_pymysql_dialect_configuration(engine: Engine) -> Callable[[DBAPIConnection, ConnectionPoolEntry], None]:
    """Creates a function to configure the PyMySQL dialect.

    Args:
        engine (Engine): The SQLAlchemy engine to configure.

    Returns:
        Callable[[DBAPIConnection, ConnectionPoolEntry], None]: A function that configures PyMySQL dialect settings.
    """
    def configure_pymysql_dialect(dbapi_connection: DBAPIConnection, connection_record: ConnectionPoolEntry):
        """Configures settings specific to PyMySQL on each connection.

        Args:
            dbapi_connection (DBAPIConnection): DB-API connection object from PyMySQL.
            connection_record (ConnectionPoolEntry): Connection record managed by SQLAlchemy.
        """
        # Disable PyMySQL's handling of ON DUPLICATE KEY UPDATE alias requirement
        engine.dialect._requires_alias_for_on_duplicate_key = False

    return configure_pymysql_dialect


def alter_session_on_oracle_connect(dbapi_connection: DBAPIConnection, connection_record: ConnectionPoolEntry):
    """SQLAlchemy event listener to alter the Oracle session settings upon connection.

    This callback function is intended to be used with SQLAlchemy's `connect` event.
    It alters the Oracle session settings by setting the `NLS_DATE_FORMAT` for the session
    to ensure that dates are formatted as 'YYYY-MM-DD HH24:MI:SS'.

    Args:
        dbapi_connection (DBAPIConnection): The raw DB-API connection object provided by the Oracle driver.
        connection_record (ConnectionPoolEntry): A record associated with the connection, managed by SQLAlchemy.
    """
    cursor = dbapi_connection.cursor()
    cursor.execute("ALTER SESSION SET NLS_DATE_FORMAT = 'YYYY-MM-DD HH24:MI:SS'")
    cursor.close()


def construct_table_reference(table: sa.Table, with_brackets: bool = False) -> str:
    """Constructs a table reference string.

    Args:
        table (sa.Table): The SQLAlchemy Table object.
        with_brackets (bool, optional): Whether to include brackets around schema and table names or not.

    Returns:
        str: The formatted table reference string.
    """
    schema_prefix = ""
    if table.schema:
        schema_prefix = f"[{table.schema}]." if with_brackets else f"{table.schema}."
    table_name = f"[{table.name}]" if with_brackets else table.name
    return f"{schema_prefix}{table_name}"



import logging
import os
from abc import ABCMeta
from contextlib import suppress
from csv import DictReader
from itertools import count, islice
from typing import AsyncGenerator, List, Optional

from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        csv_file = self.properties["file"]

        if os.path.isabs(csv_file) or context is None:
            self.file = csv_file
        else:
            self.file = os.path.join(context.properties.get("data_path"), csv_file)

        logger.debug(f"file: {self.file}")

        self.encoding = self.properties.get("encoding", "utf-8")
        self.batch_size = self.properties.get("batch_size", 1000)
        self.fields = self.properties.get("fields")
        self.skip = self.properties.get("skip", 0)
        self.delimiter = self.properties.get("delimiter", ",")
        self.quotechar = self.properties.get("quotechar", "\"")

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        logger.debug("Reading CSV")

        with open(self.file, "r", encoding=self.encoding) as read_obj:
            reader = DictReader(read_obj, fieldnames=self.fields, delimiter=self.delimiter, quotechar=self.quotechar)
            counter = iter(count())

            for _ in range(self.skip):
                with suppress(StopIteration):
                    next(reader)

            while True:
                sliced = islice(reader, self.batch_size)
                records = [{self.MSG_ID_FIELD: f"{next(counter)}", **record} for record in sliced]

                if not records:
                    logger.debug(f"Done reading {self.file}")
                    return

                logger.debug(f"Producing {len(records)} records")

                yield records



from __future__ import annotations

import importlib
import logging
import os
import sys
from abc import ABCMeta, abstractmethod
from os import path
from typing import Any, Dict, List, Optional

from datayoga_core import utils
from datayoga_core.context import Context
from datayoga_core.result import BlockResult
from jsonschema import validate

logger = logging.getLogger("dy")


class Block(metaclass=ABCMeta):
    INTERNAL_FIELD_PREFIX = "__$$"
    MSG_ID_FIELD = f"{INTERNAL_FIELD_PREFIX}msg_id"
    RESULT_FIELD = f"{INTERNAL_FIELD_PREFIX}result"
    OPCODE_FIELD = f"{INTERNAL_FIELD_PREFIX}opcode"
    """Block.

    Attributes:
        properties Dict[str, Any]: Block properties.
    """

    def __init__(self, properties: Optional[Dict[str, Any]] = None):
        """Constructs a block.

        Args:
            properties (Optional[Dict[str, Any]]): Block [properties].
        """
        self.properties = properties or {}
        self.validate()

    def validate(self):
        """Validates block against its JSON Schema"""
        logger.debug(f"validating {self.properties}")
        validate(instance=self.properties, schema=self.get_json_schema())

    def get_json_schema(self) -> Dict[str, Any]:
        """Returns the JSON Schema for this block.

        Returns:
            Dict[str, Any]: JSON Schema.
        """
        json_schema_file = path.join(
            utils.get_bundled_dir(),
            os.path.relpath(
                os.path.dirname(sys.modules[self.__module__].__file__),
                start=os.path.dirname(__file__)),
            "block.schema.json") if utils.is_bundled() else path.join(
            os.path.dirname(os.path.realpath(sys.modules[self.__module__].__file__)),
            "block.schema.json")
        logger.debug(f"loading schema from {json_schema_file}")
        return utils.read_json(json_schema_file)

    @abstractmethod
    def init(self, context: Optional[Context] = None):
        """Initializes block.

        Args:
            context (Context, optional): Context. Defaults to None.
        """
        raise NotImplementedError

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        """Transforms data.

        Args:
            data (List[Dict[str, Any]]): Data.

        Returns:
            BlockResult: Block result.
        """
        pass

    def stop(self):
        """Cleans the block connections and state."""
        pass

    def get_block_name(self):
        return os.path.basename(os.path.dirname(sys.modules[self.__module__].__file__))

    @staticmethod
    def create(block_name: str, properties: Dict[str, Any]) -> Block:
        module_name = f"datayoga_core.blocks.{block_name}.block"
        module = importlib.import_module(module_name)
        block: Block = getattr(module, "Block")(properties)
        return block



from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class Context:
    """
    Context

    Attributes:
        properties Dict[str, Any]: Context properties
        state (Dict[str, Any]): Context state
    """
    properties: Optional[Dict[str, Any]] = None
    state: Optional[Dict[str, Any]] = None



import asyncio
import logging
from typing import AsyncGenerator, List, Optional

import orjson
from azure.eventhub import EventData, PartitionContext
from azure.eventhub.aio import EventHubConsumerClient
from azure.eventhub.extensions.checkpointstoreblobaio import \
    BlobCheckpointStore
from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer):
    """Azure Event Hub block for reading events."""

    def init(self, context: Optional[Context] = None):
        """Initializes the block.

        Args:
            context (Context, optional): The block context. Defaults to None.
        """
        logger.debug(f"Initializing {self.get_block_name()}")

        self.batch_size = self.properties.get("batch_size", 300)

        self.consumer_client = EventHubConsumerClient.from_connection_string(
            conn_str=self.properties["event_hub_connection_string"],
            consumer_group=self.properties["event_hub_consumer_group_name"],
            eventhub_name=self.properties["event_hub_name"],
            checkpoint_store=BlobCheckpointStore.from_connection_string(
                self.properties["checkpoint_store_connection_string"],
                self.properties["checkpoint_store_container_name"])
        )

        self.events = {}  # Retrieved events by sequence number, used for acknowledging them once processed
        self.messages = asyncio.Queue()

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        """Starts the event receiving process and yield batches of messages.

        Yields:
            AsyncGenerator[List[Message], None]: A generator of message batches.
        """
        logger.debug(f"Running {self.get_block_name()}")

        logger.debug("Starting event receiving process")
        asyncio.create_task(self.receive_batch())

        while True:
            if not self.messages.empty():
                batch = []
                while not self.messages.empty():
                    message = await self.messages.get()
                    batch.append(message)

                yield batch

            await asyncio.sleep(0.1)

    async def receive_batch(self):
        """Receives events in batches from the Event Hub."""
        await self.consumer_client.receive_batch(
            on_event_batch=self.on_event_batch,
            max_batch_size=self.batch_size,
            starting_position="-1",  # read from the beginning of the partition.
        )

    async def on_event_batch(self, partition_context: PartitionContext, events: List[EventData]):
        """Processes each batch of events received from the Event Hub.

        Args:
            partition_context (PartitionContext): The partition context.
            events (List[EventData]): The list of events in the batch.
        """
        logger.debug(f"Received batch of events from partition: {partition_context.partition_id}")

        for event in events:
            try:
                payload = orjson.loads(event.body_as_str(encoding="UTF-8"))
                msg_id = event.system_properties[b"x-opt-sequence-number"]
                self.events[msg_id] = (event, partition_context)
                payload[self.MSG_ID_FIELD] = msg_id
                await self.messages.put(payload)
            except Exception as e:
                logger.error(e)

    async def complete_events(self, msg_ids: List[str]):
        """Completes the events and update the checkpoint.

        Args:
            msg_ids (List[str]): The list of message IDs to complete.
        """
        for msg_id in msg_ids:
            logger.debug(f"Acking {msg_id} event")
            event, partition_context = self.events.pop(msg_id, (None, None))

            if event is not None:
                await partition_context.update_checkpoint(event)
            else:
                logger.warning(f"Couldn't find event {msg_id} for acknowledging")

    def ack(self, msg_ids: List[str]):
        """Acknowledges the completion of events.

        Args:
            msg_ids (List[str]): The list of message IDs to acknowledge.
        """
        asyncio.create_task(self.complete_events(msg_ids))
